/*----- PROTECTED REGION ID(AlarmNotify.cpp) ENABLED START -----*/
//=============================================================================
//
// file :        AlarmNotify.cpp
//
// description : C++ source for the AlarmNotify class and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               AlarmNotify are implemented in this file.
//
// project :     Alarm Notify
//
// This file is part of Tango device class.
// 
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
// 
//
// Elettra-Sincrotrone Trieste S.C.p.A. di interesse nazionale
// Strada Statale 14 - km 163,5 in AREA Science Park
// 34149 Basovizza, Trieste ITALY
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include <AlarmNotify.h>
#include <AlarmNotifyClass.h>
#include <netdb.h> //for getaddrinfo
#include <string>
#include <unordered_map>
#include "PushThread.h"
#include "SubscribeThread.h"

/*----- PROTECTED REGION END -----*/	//	AlarmNotify.cpp

/**
 *  AlarmNotify class description:
 *    Alarm handler notifier
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name  |  Method name
//================================================================
//  State         |  Inherited (no method)
//  Status        |  Inherited (no method)
//  GetAlarmInfo  |  get_alarm_info
//================================================================

//================================================================
//  Attributes managed are:
//================================================================
//  alarmStatus     |  Tango::DevString	Spectrum  ( max = 10000)
//  handlerStatus   |  Tango::DevString	Spectrum  ( max = 1000)
//  alarmReceivers  |  Tango::DevString	Spectrum  ( max = 10000)
//================================================================

namespace AlarmNotify_ns
{
/*----- PROTECTED REGION ID(AlarmNotify::namespace_starting) ENABLED START -----*/

//	static initializations
typedef std::vector<string> Vec;
struct Contained
{
    Vec &_sequence;
    Vec _sequence2;
    Contained(Vec &vec) : _sequence(vec){
    	Vec _sequence_upper(vec);
    	for(Vec::iterator it = _sequence_upper.begin(); it!=_sequence_upper.end(); it++)
    	{
    		string result("");
   			string::size_type found;
   			string separator(":");
   			found = it->find_first_of(separator);
    		if(found != string::npos) {
    			if(found > 0) {
    				result = it->substr(0,found);
    			}
    		}
    		else if(it->length() > 0) {
    			result = *it;
    		}
    		std::transform(result.begin(), result.end(), result.begin(), (int(*)(int))toupper);
    		_sequence2.push_back(result);
    	}
    }
    bool operator()(string i) const
    {
		string result("");
		string::size_type found;
		string separator(":");
		found = i.find_first_of(separator);
		if(found != string::npos) {
			if(found > 0) {
				result = i.substr(0,found);
			}
		}
		else if(i.length() > 0) {
			result = i;
		}
    	std::transform(result.begin(), result.end(), result.begin(), (int(*)(int))toupper);
        return _sequence2.end() != std::find(_sequence2.begin(), _sequence2.end(), result);
    }
};

//	static initialization
map<string, string> AlarmStringUtils::domain_map;
map<Tango::DevState, string> AlarmStringUtils::tango_states ={ 
	{ Tango::ON, "ON" },
	{ Tango::OFF, "OFF" },
	{ Tango::CLOSE, "CLOSE" },
	{ Tango::OPEN, "OPEN" },
	{ Tango::INSERT, "INSERT" },
	{ Tango::EXTRACT, "EXTRACT" },
	{ Tango::MOVING, "MOVING" },
	{ Tango::STANDBY, "STANDBY" },
	{ Tango::FAULT, "FAULT" },
	{ Tango::INIT, "INIT" },
	{ Tango::RUNNING, "RUNNING" },
	{ Tango::ALARM, "ALARM" },
	{ Tango::DISABLE, "DISABLE" },
	{ Tango::UNKNOWN, "UNKNOWN" }
};

vector<string> AlarmStringUtils::alarm_states ={"ACKED","ERROR","NORM","OOSRV","RTNUN","SHLVD","UNACK"};

void AlarmStringUtils::string_explode(string str, const string& separator, vector<string>& results)
{
	const char* t = " \t\n\r\f\v";
	auto found = str.find_first_of(separator);
	while(found != string::npos)
	{
		if(found > 0)
		{
			string trim = str.substr(0,found);
            trim.erase(0, trim.find_first_not_of(t));
            trim.erase(trim.find_last_not_of(t) + 1);
			results.push_back(trim);
		}
		str = str.substr(found+1);
		found = str.find_first_of(separator);
	}
	if(str.length() > 0)
	{
		string trim = str;
		trim.erase(0, trim.find_first_not_of(t));
		trim.erase(trim.find_last_not_of(t) + 1);
		results.push_back(trim);
	}
}

void AlarmStringUtils::string_vector2map(const vector<string> &str, const string &separator, unordered_map<string,string> &results)
{
	for(const auto &it : str)
	{
		auto found_eq = it.find_first_of(separator);
		if(found_eq != string::npos && found_eq > 0)
		{
			results.insert(make_pair(it.substr(0,found_eq),it.substr(found_eq+1)));
		}
	}
}

void AlarmStringUtils::fix_tango_host(string &attr)
{
	std::transform(attr.begin(), attr.end(), attr.begin(), (int(*)(int))tolower);		//transform to lowercase
	string::size_type	start = attr.find("tango://");
	//if not fqdn, add TANGO_HOST
	if (start == string::npos)
	{
		//TODO: get from device/class/global property
		char	*env = getenv("TANGO_HOST");
		if (env==NULL)
		{
			return;
		}
		else
		{
			string	s(env);
			add_domain(s);
			attr = string("tango://") + s + "/" + attr;
			return;
		}
	}
	string facility = get_only_tango_host(attr);
	add_domain(facility);
	string attr_name = get_only_attr_name(attr);
	attr = string("tango://")+ facility + string("/") + attr_name;
}

void AlarmStringUtils::add_domain(string &str)
{
	string::size_type	end1 = str.find(".");
	if (end1 == string::npos)
	{
		//get host name without tango://
		string::size_type	start = str.find("tango://");
		if (start == string::npos)
		{
			start = 0;
		}
		else
		{
			start = 8;	//tango:// len
		}
		string::size_type	end2 = str.find(":", start);

		string th = str.substr(start, end2);
		string with_domain = str;

		map<string,string>::iterator it_domain = domain_map.find(th);
		if(it_domain != domain_map.end())
		{
			with_domain = it_domain->second;
			//cout << __func__ <<": found domain in map -> " << with_domain<<endl;
			str = with_domain;
			return;
		}

		struct addrinfo hints;
//		hints.ai_family = AF_INET; // use AF_INET6 to force IPv6
//		hints.ai_flags = AI_CANONNAME|AI_CANONIDN;
		memset(&hints, 0, sizeof hints);
		hints.ai_family = AF_UNSPEC; /*either IPV4 or IPV6*/
		hints.ai_socktype = SOCK_STREAM;
		hints.ai_flags = AI_CANONNAME;
		struct addrinfo *result;
		int ret = getaddrinfo(th.c_str(), NULL, &hints, &result);
		if (ret != 0)
		{
			cout << __func__<< ": getaddrinfo error='" << gai_strerror(ret)<<"' while looking for " << th<<endl;
			return;
		}
		if(result == NULL)
		{
			cout << __func__ << ": getaddrinfo did not return domain information for " 
			     << th << " (result == NULL)" << endl;
			return;
		}
		if (result->ai_canonname == NULL)
		{
			cout << __func__ << ": getaddrinfo did not return domain information for " 
			     << th << " (result->ai_canonname == NULL)" << endl;
			freeaddrinfo(result);
			return;
		}
		with_domain = string(result->ai_canonname) + str.substr(end2);
		//cout << __func__ <<": found domain -> " << with_domain<<endl;
		domain_map.insert(make_pair(th, with_domain));
		freeaddrinfo(result); // all done with this structure
		str = with_domain;
		return;
	}
	else
	{
		return;
	}
}

string AlarmStringUtils::get_only_attr_name(const string &str)
{
	string::size_type	start = str.find("tango://");
	if (start == string::npos)
		return str;
	else
	{
		start += 8; //	"tango://" length
		start = str.find('/', start);
		start++;
		string	tagname = str.substr(start);
		return tagname;
	}
}

string AlarmStringUtils::get_only_tango_host(const string &str)
{
	string::size_type	start = str.find("tango://");
	if (start == string::npos)
	{
		return "unknown";
	}
	else
	{
		start += 8; //	"tango://" length
		string::size_type	end = str.find('/', start);
		string th = str.substr(start, end-start);
		return th;
	}
}

void AlarmStringUtils::parse_tango_host(const string &th, string &host, int &port)
{
	string::size_type	end = th.find(':');
	host = th.substr(0, end);
	string ports=th.substr(end+1);
	port = stoi(ports);
}

string AlarmStringUtils::remove_domain(const string &str)
{
	string::size_type	end1 = str.find(".");
	if (end1 == string::npos)
	{
		return str;
	}
	else
	{
		string::size_type	start = str.find("tango://");
		if (start == string::npos)
		{
			start = 0;
		}
		else
		{
			start = 8;	//tango:// len
		}
		string::size_type	end2 = str.find(":", start);
		if(end1 > end2)	//'.' not in the tango host part
			return str;
		string th = str.substr(0, end1);
		th += str.substr(end2, str.size()-end2);
		return th;
	}
}

void AlarmStringUtils::split_tag_handler(const string &name, string &handler, string &tag)
{
	auto pos = name.find_last_of("/");
	if (pos != string::npos)
	{
		handler=name.substr(0,pos);
		tag=name.substr(pos+1,name.size()-pos-1);
	}
}

auto AlarmStringUtils::compare_without_domain(const string &str1, const string &str2) -> bool
{
        string str1_nd = remove_domain(str1);
        string str2_nd = remove_domain(str2);
        return (str1_nd==str2_nd);
}
//NORM,RTNUN: ab@c,de@f;UNACK,ACKED: ab@c,de@f; OOSRV,SHLVD: op@e; ERROR: ctrl@e
void AlarmStringUtils::parse_receivers(const string &receivers_str, state_addr_map_t &st_receivers)
{
	vector<string> st_rec_list;
	AlarmStringUtils::string_explode(receivers_str, ";", st_rec_list);
	unordered_map<string,string> st_rec_map;
	AlarmStringUtils::string_vector2map(st_rec_list, ":", st_rec_map);
	for(auto it : st_rec_map)
	{
		vector<string> states;
		AlarmStringUtils::string_explode(it.first, ",", states);
		vector<string> dest;
		AlarmStringUtils::string_explode(it.second, ",", dest);
		for(auto st: states)
		{
			receivers_t tmp{dest};
#if 0
			if(!dest.empty())
				cout << __func__ << ": inserting map state="<<st<<" -> dest="<<dest[0]<<" dest.size="<<dest.size()<<endl;
#endif
			st_receivers.insert(make_pair(st,tmp));
		}
	}
}
void AlarmStringUtils::print_receivers(string &receivers_str, const state_addr_map_t &st_receivers)
{
	for(auto st : st_receivers)
	{
		ostringstream out;
		out << st.first << ":";
		std::copy(
			st.second.emails.begin(),
			st.second.emails.end(),
			std::ostream_iterator<string>(out, ",")
		);
		receivers_str += out.str();
		receivers_str.erase(receivers_str.find_last_not_of(",") + 1);//delete last ','
		receivers_str += ";";
	}
	receivers_str.erase(receivers_str.find_last_not_of(",;") + 1);//delete last ',;'
}
//return false if str1 >= str2
bool compare_tango_names(string str1, string str2)
{
	//cout << __func__<< ": entering with '" << str1<<"' - '" << str2<<"'" << endl;
	if(str1 == str2)
	{
		//cout << __func__<< ": EQUAL 1 -> '" << str1<<"'=='" << str2<<"'" << endl;
		return false;
	}
	AlarmStringUtils::fix_tango_host(str1);
	AlarmStringUtils::fix_tango_host(str2);
	if(str1 == str2)
	{
		//cout << __func__<< ": EQUAL 2 -> '" << str1<<"'=='" << str2<<"'" << endl;
		return false;
	}
	string str1_nd = AlarmStringUtils::remove_domain(str1);
	string str2_nd = AlarmStringUtils::remove_domain(str2);
	if(str1_nd == str2_nd)
	{
		//cout << __func__<< ": EQUAL 3 -> '" << str1_nd<<"'=='" << str2_nd<<"'" << endl;
		return false;
	}
	bool result=str1_nd<str2_nd;
	//cout << __func__<< ": DIFFERENTS -> '" << str1_nd<< (result ? "'<'" : "'>'") << str2_nd<<"' result="<<result << endl;
	return result;
}
#if 0
template<typename T, typename... Args>
std::unique_ptr<T> make_unique(Args&&... args)
{
    return std::unique_ptr<T>(new T(std::forward<Args>(args)...));
}
#endif

/*----- PROTECTED REGION END -----*/	//	AlarmNotify::namespace_starting

//--------------------------------------------------------
/**
 *	Method      : AlarmNotify::AlarmNotify()
 *	Description : Constructors for a Tango device
 *                implementing the classAlarmNotify
 */
//--------------------------------------------------------
AlarmNotify::AlarmNotify(Tango::DeviceClass *cl, string &s)
 : TANGO_BASE_CLASS(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(AlarmNotify::constructor_1) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::constructor_1
}
//--------------------------------------------------------
AlarmNotify::AlarmNotify(Tango::DeviceClass *cl, const char *s)
 : TANGO_BASE_CLASS(cl, s)
{
	/*----- PROTECTED REGION ID(AlarmNotify::constructor_2) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::constructor_2
}
//--------------------------------------------------------
AlarmNotify::AlarmNotify(Tango::DeviceClass *cl, const char *s, const char *d)
 : TANGO_BASE_CLASS(cl, s, d)
{
	/*----- PROTECTED REGION ID(AlarmNotify::constructor_3) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::constructor_3
}

//--------------------------------------------------------
/**
 *	Method      : AlarmNotify::delete_device()
 *	Description : will be called at device destruction or at init command
 */
//--------------------------------------------------------
void AlarmNotify::delete_device()
{
	DEBUG_STREAM << "AlarmNotify::delete_device() " << device_name << endl;
	/*----- PROTECTED REGION ID(AlarmNotify::delete_device) ENABLED START -----*/
	
	//	Delete device allocated objects
	push_thread->abort();
	DEBUG_STREAM << "       Push thread Stopped " << endl;
	push_thread->join(nullptr);
	DEBUG_STREAM << "	Stopping subscribe thread" << endl;
	shared->stop_thread();
	DEBUG_STREAM << "	Subscribe thread Stopped " << endl;
	thread->join(nullptr);

	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::delete_device
	delete[] attr_alarmStatus_read;
	delete[] attr_handlerStatus_read;
	delete[] attr_alarmReceivers_read;
}

//--------------------------------------------------------
/**
 *	Method      : AlarmNotify::init_device()
 *	Description : will be called at device initialization.
 */
//--------------------------------------------------------
void AlarmNotify::init_device()
{
	DEBUG_STREAM << "AlarmNotify::init_device() create device " << device_name << endl;
	/*----- PROTECTED REGION ID(AlarmNotify::init_device_before) ENABLED START -----*/
	
	//	Initialization before get_device_property() call
	attr_alarmStatus_read = NULL;
	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::init_device_before
	

	//	Get the device properties from database
	get_device_property();
	
	attr_alarmStatus_read = new Tango::DevString[10000];
	attr_handlerStatus_read = new Tango::DevString[1000];
	attr_alarmReceivers_read = new Tango::DevString[10000];
	/*----- PROTECTED REGION ID(AlarmNotify::init_device) ENABLED START -----*/
	
	//	Initialize device
	for(auto it : propertyList)
	{
		vector<string> prop_name;
		AlarmStringUtils::string_explode(it, "/", prop_name);
		if(prop_name.size() != 2)
		{
			DEBUG_STREAM << __func__ << ": BAD property " << it << " size=" << prop_name.size();
			continue;
		}
		vector<string> hlist;
		auto db = make_unique<Tango::Database>();
		try
		{
			Tango::DbData db_data;
			db_data.push_back((Tango::DbDatum(prop_name[1])));
			db->get_property(prop_name[0],db_data);
			db_data[0] >> hlist;
		}catch(Tango::DevFailed &e)
		{
			ERROR_STREAM << __FUNCTION__ << " Error reading Database property='" << e.errors[0].desc << "'";
		}
		for(auto hit : hlist)
		{
			if (std::find(handlerList.begin(), handlerList.end(), hit) == handlerList.end())
			{
				handlerList.push_back(hit);
			}
		}
	}
	handler_t tmp;
	for(auto it : handlerList)
	{
		try
		{
			//tmp.dp = std::unique_ptr<Tango::DeviceProxy>(new Tango::DeviceProxy(it));
			tmp.dp = make_shared<Tango::DeviceProxy>(it);
		}
		catch(Tango::DevFailed &e)
		{
			//tmp.dp = std::unique_ptr<Tango::DeviceProxy>(nullptr);
			tmp.dp = nullptr;
		}
		string hndlname(it);
		AlarmStringUtils::fix_tango_host(hndlname);
		DEBUG_STREAM << __func__<<": adding handler:"<<hndlname;
		handler_list_fix.push_back(hndlname);
		handlerMap.insert(make_pair(hndlname,std::move(tmp)));
	}
	alarm_receivers = make_unique<Receivers>(this);
//	get_alarm_list_db(alarm_list, receiverMap);
	AlarmStringUtils::parse_receivers(receivers,globalReceiverMap);
	DEBUG_STREAM << __func__ << ": parsed global receivers, map size=" << globalReceiverMap.size() << endl;
	for(auto gr : groupReceivers)
	{
		auto found = gr.find_first_of("=");
		if(found != string::npos && found > 0)
		{
			state_addr_map_t rec;
			AlarmStringUtils::parse_receivers(gr.substr(found+1),rec);
			DEBUG_STREAM << __func__ << ": parsed group " << gr.substr(found+1) << " receivers, map size=" << rec.size() << endl;
			groupReceiverMap.insert(make_pair(gr.substr(0,found),rec));
		}
	}
	DEBUG_STREAM << __func__ << ": parsed group receivers, map size=" << groupReceiverMap.size() << endl;

	shared = std::make_shared<SharedData>(this);
	thread = std::unique_ptr<SubscribeThread, std::function<void(SubscribeThread*)>>(new SubscribeThread(this)
			, [](SubscribeThread* /*unused*/){});
	push_thread = std::unique_ptr<PushThread, std::function<void(PushThread*)>>(
		new PushThread(this
			, Tango::Util::instance()->get_ds_inst_name()
			, configuration)
		, [](PushThread* /*unused*/){});

	push_thread->start();
	thread->start();
	shared->wait_initialized();
#if 0
	for(auto it : alarm_list)
	{
		try
		{
			shared->add(it);
		}
		catch(Tango::DevFailed &e)
		{
			INFO_STREAM << ": error adding " << it << ", err=" << e.errors[0].desc;
		}
	}
	shared->start_all();
#endif
	


	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::init_device
}

//--------------------------------------------------------
/**
 *	Method      : AlarmNotify::get_device_property()
 *	Description : Read database to initialize property data members.
 */
//--------------------------------------------------------
void AlarmNotify::get_device_property()
{
	/*----- PROTECTED REGION ID(AlarmNotify::get_device_property_before) ENABLED START -----*/
	
	//	Initialize property data members
	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::get_device_property_before


	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("HandlerList"));
	dev_prop.push_back(Tango::DbDatum("PropertyList"));
	dev_prop.push_back(Tango::DbDatum("MaxSearchSize"));
	dev_prop.push_back(Tango::DbDatum("Configuration"));
	dev_prop.push_back(Tango::DbDatum("Receivers"));
	dev_prop.push_back(Tango::DbDatum("SubscribePeriod"));
	dev_prop.push_back(Tango::DbDatum("GroupReceivers"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);
	
		//	get instance on AlarmNotifyClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		AlarmNotifyClass	*ds_class =
			(static_cast<AlarmNotifyClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize HandlerList from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  handlerList;
		else {
			//	Try to initialize HandlerList from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  handlerList;
		}
		//	And try to extract HandlerList value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  handlerList;

		//	Try to initialize PropertyList from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  propertyList;
		else {
			//	Try to initialize PropertyList from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  propertyList;
		}
		//	And try to extract PropertyList value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  propertyList;

		//	Try to initialize MaxSearchSize from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  maxSearchSize;
		else {
			//	Try to initialize MaxSearchSize from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  maxSearchSize;
		}
		//	And try to extract MaxSearchSize value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  maxSearchSize;

		//	Try to initialize Configuration from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  configuration;
		else {
			//	Try to initialize Configuration from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  configuration;
		}
		//	And try to extract Configuration value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  configuration;

		//	Try to initialize Receivers from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  receivers;
		else {
			//	Try to initialize Receivers from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  receivers;
		}
		//	And try to extract Receivers value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  receivers;

		//	Try to initialize SubscribePeriod from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  subscribePeriod;
		else {
			//	Try to initialize SubscribePeriod from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  subscribePeriod;
		}
		//	And try to extract SubscribePeriod value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  subscribePeriod;

		//	Try to initialize GroupReceivers from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  groupReceivers;
		else {
			//	Try to initialize GroupReceivers from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  groupReceivers;
		}
		//	And try to extract GroupReceivers value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  groupReceivers;

	}

	/*----- PROTECTED REGION ID(AlarmNotify::get_device_property_after) ENABLED START -----*/
	
	//	Check device property data members init
	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::get_device_property_after
}

//--------------------------------------------------------
/**
 *	Method      : AlarmNotify::always_executed_hook()
 *	Description : method always executed before any command is executed
 */
//--------------------------------------------------------
void AlarmNotify::always_executed_hook()
{
	DEBUG_STREAM << "AlarmNotify::always_executed_hook()  " << device_name << endl;
	/*----- PROTECTED REGION ID(AlarmNotify::always_executed_hook) ENABLED START -----*/
	
	//	code always executed before all requests
	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method      : AlarmNotify::read_attr_hardware()
 *	Description : Hardware acquisition for attributes
 */
//--------------------------------------------------------
void AlarmNotify::read_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "AlarmNotify::read_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(AlarmNotify::read_attr_hardware) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::read_attr_hardware
}

//--------------------------------------------------------
/**
 *	Read attribute alarmStatus related method
 *	Description: List of:
 *               alarm name - status - notification counter
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Spectrum max = 10000
 */
//--------------------------------------------------------
void AlarmNotify::read_alarmStatus(Tango::Attribute &attr)
{
	DEBUG_STREAM << "AlarmNotify::read_alarmStatus(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(AlarmNotify::read_alarmStatus) ENABLED START -----*/

	size_t i=0;
	vector<string> al_list;
	shared->get_sig_list_status(al_list);
	DEBUG_STREAM << __func__ << ": alarm list size=" << al_list.size();
	for(const auto &it : al_list)
	{
		attr_alarmStatus_read[i++] = CORBA::string_dup(it.c_str());
	}
	//	Set the attribute value
	attr.set_value(attr_alarmStatus_read, i);
	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::read_alarmStatus
}
//--------------------------------------------------------
/**
 *	Read attribute handlerStatus related method
 *	Description: List of:
 *               Handler Device Name = Tango State
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Spectrum max = 1000
 */
//--------------------------------------------------------
void AlarmNotify::read_handlerStatus(Tango::Attribute &attr)
{
	DEBUG_STREAM << "AlarmNotify::read_handlerStatus(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(AlarmNotify::read_handlerStatus) ENABLED START -----*/
	size_t i=0;
	for(auto it = handlerMap.begin(); it != handlerMap.end(); it++)
	{
		string hstatus=it->first+": ";
		Tango::DevState hstate;
		try
		{
			if(it->second.dp)
			{
				it->second.dp->read_attribute("State") >> hstate;
				hstatus+=AlarmStringUtils::tango_states.at(hstate);
			}
			else
			{
				hstatus+="NOT CONNECTED";
			}
		}
		catch(Tango::DevFailed &e)
		{
			INFO_STREAM << __func__<<": unable to read State from " << it->first;
			hstatus+=e.errors[0].desc;
		}
		catch(const std::out_of_range& e)
		{
			INFO_STREAM << "Unknown state: " << hstate;
			hstatus+=to_string(static_cast<int>(hstate));
		}
		attr_handlerStatus_read[i++] = CORBA::string_dup(hstatus.c_str());
	}
	//	Set the attribute value
	attr.set_value(attr_handlerStatus_read, i);
	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::read_handlerStatus
}
//--------------------------------------------------------
/**
 *	Read attribute alarmReceivers related method
 *	Description: List of:
 *               alarm name - alarm states: receivers; alarm state 2: receivers
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Spectrum max = 10000
 */
//--------------------------------------------------------
void AlarmNotify::read_alarmReceivers(Tango::Attribute &attr)
{
	DEBUG_STREAM << "AlarmNotify::read_alarmReceivers(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(AlarmNotify::read_alarmReceivers) ENABLED START -----*/
	size_t i=0;
	vector<string> al_list;
	shared->get_sig_list(al_list);
	DEBUG_STREAM << __func__ << ": alarm list size=" << al_list.size();
	for(const auto &it : al_list)
	{
		string dest;
		string groups;
		state_addr_map_t rec;
		//alarm_receivers->print_receivers(it, dest);
		try
		{
			rec =  alarm_receivers->get_state_addr(it);
		}
		catch(const std::out_of_range& ooe)
		{
			DEBUG_STREAM << __func__ << ": no receivers for " << it;
			continue;
		}
		try
		{
			groups = alarm_receivers->get_groups(it);
		}
		catch(const std::out_of_range& ooe)
		{
			DEBUG_STREAM << __func__ << ": no group for " << it;
			continue;
		}		

		for(auto st : AlarmStringUtils::alarm_states)
		{
			string receivers;
            prepare_receivers(it, st, groups, receivers);
			if(!receivers.empty())
			{
				dest += st + ":" + receivers + ";";
			}
		}
		dest.erase(dest.find_last_not_of(";") + 1);//delete last ';'
		attr_alarmReceivers_read[i++] = CORBA::string_dup((it+" - " + dest).c_str());
	}
	//	Set the attribute value
	attr.set_value(attr_alarmReceivers_read, i);
	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::read_alarmReceivers
}

//--------------------------------------------------------
/**
 *	Method      : AlarmNotify::add_dynamic_attributes()
 *	Description : Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void AlarmNotify::add_dynamic_attributes()
{
	/*----- PROTECTED REGION ID(AlarmNotify::add_dynamic_attributes) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic attributes if any
	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command GetAlarmInfo related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarStringArray *AlarmNotify::get_alarm_info(Tango::DevString argin)
{
	Tango::DevVarStringArray *argout;
	DEBUG_STREAM << "AlarmNotify::GetAlarmInfo()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(AlarmNotify::get_alarm_info) ENABLED START -----*/
	
	//	Add your own code

	auto signal = shared->get_signal(argin);
	argout = new Tango::DevVarStringArray();

	argout->length(8);
	ostringstream st;
	string state;
	st << "Name:        " << argin;
	(*argout)[0] = CORBA::string_dup(st.str().c_str());
	st.str("");
	st << "Event state: " << (signal->is_not_subscribed() ? "NOT CONNECTED" : "SUBSCRIBED");
	(*argout)[1] = CORBA::string_dup(st.str().c_str());
	st.str("");
	st << "Event error: " << signal->get_error();
	(*argout)[2] = CORBA::string_dup(st.str().c_str());
	st.str("");
	state = signal->get_alarm_state();
	st << "Alarm State: " << state;
	(*argout)[3] = CORBA::string_dup(st.str().c_str());
	st.str("");
	st << "Counter:     " << signal->get_ok_db();
	(*argout)[4] = CORBA::string_dup(st.str().c_str());
	string dest;
	alarm_receivers->print_receivers(argin, dest);
	st.str("");
	st << "Dest:        " << dest;
	(*argout)[5] = CORBA::string_dup(st.str().c_str());
	st.str("");
	st << "Global Dest: " << receivers;
	(*argout)[6] = CORBA::string_dup(st.str().c_str());
	st.str("");
	st << "Group Dest:  ";
	string groups;
	vector<string> grdest;
	try
	{
		groups = alarm_receivers->get_groups(argin);
	}
	catch(const std::out_of_range& ooe)
	{
		DEBUG_STREAM << __func__ << ": no group for " << argin;
	}
	
	if(!groups.empty())
	{
		vector<string> v_groups;
		AlarmStringUtils::string_explode(groups, "|", v_groups);
		for(const auto &gr : v_groups)
		{
			state_addr_map_t rec;
			string gr_rec;
			try
			{
				rec = groupReceiverMap.at(gr);
				AlarmStringUtils::print_receivers(gr_rec, rec);
			}
			catch(const std::out_of_range& ooe)
			{
				//DEBUG_STREAM << __func__ << ": no group receiver for " << gr;
				continue;
			}
			try
			{
				grdest = rec.at(state).emails;
			}
			catch(const std::out_of_range& ooe)
			{
				//DEBUG_STREAM << __func__ << ": no group receiver for " << argin << ", state="<<state<<", group=" << gr << " err=" << ooe.what();
				continue;
			}
			if(!gr_rec.empty())
			{
				st<<gr_rec<<"|";
			}
		}
	}
	string trim(st.str());
	trim.erase(trim.find_last_not_of(",|") + 1);//delete last ',!'
	(*argout)[7] = CORBA::string_dup(trim.c_str());
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::get_alarm_info
	return argout;
}
//--------------------------------------------------------
/**
 *	Method      : AlarmNotify::add_dynamic_commands()
 *	Description : Create the dynamic commands if any
 *                for specified device.
 */
//--------------------------------------------------------
void AlarmNotify::add_dynamic_commands()
{
	/*----- PROTECTED REGION ID(AlarmNotify::add_dynamic_commands) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic commands if any
	
	/*----- PROTECTED REGION END -----*/	//	AlarmNotify::add_dynamic_commands
}

/*----- PROTECTED REGION ID(AlarmNotify::namespace_ending) ENABLED START -----*/

//	Additional Methods
void AlarmNotify::prepare_receivers(const string &attr_name, const string &state, const string &groups, string &receivers)
{
	vector<string> dest;
	vector<string> gldest;
	vector<string> grdest;
	try
	{//TODO: get state from event value!!
		dest = alarm_receivers->get_receivers(attr_name, state).emails;
	}
	catch(const std::out_of_range& ooe)
	{
		//DEBUG_STREAM << __func__ << ": no receiver for " << attr_name << ", state="<<state<<" err=" << ooe.what();
	}
	try
	{
		gldest = globalReceiverMap.at(state).emails;
	}
	catch(const std::out_of_range& ooe)
	{
		//DEBUG_STREAM << __func__ << ": no global receiver for " << attr_name << ", state="<<state<<" err=" << ooe.what();
	}
	if(!gldest.empty())
	{
		dest.insert(
			dest.end(),
			std::make_move_iterator(gldest.begin()),
			std::make_move_iterator(gldest.end())
			);
	}
	if(!groups.empty())
	{
		vector<string> v_groups;
		AlarmStringUtils::string_explode(groups, "|", v_groups);
		for(const auto &gr : v_groups)
		{
			state_addr_map_t rec;
			try
			{
				rec = groupReceiverMap.at(gr);
			}
			catch(const std::out_of_range& ooe)
			{
				//DEBUG_STREAM << __func__ << ": no group receiver for " << gr;
				continue;
			}
			try
			{
				grdest = rec.at(state).emails;
			}
			catch(const std::out_of_range& ooe)
			{
				//DEBUG_STREAM << __func__ << ": no group receiver for " << attr_name << ", state="<<state<<", group=" << gr << " err=" << ooe.what();
				continue;
			}
			if(!grdest.empty())
			{
				dest.insert(
					dest.end(),
					std::make_move_iterator(grdest.begin()),
					std::make_move_iterator(grdest.end())
					);
			}
		} 
	}
	if(!dest.empty())
	{
		//sort and remove duplicates
		sort(dest.begin(), dest.end());
		vector<string>::iterator new_end = \
			unique(dest.begin(), dest.end());
		while (dest.end() != new_end) 
			dest.pop_back();

		stringstream tmp_rec;
		// Print out concatenated vector.
		std::copy(
			dest.begin(),
			dest.end(),
			std::ostream_iterator<string>(tmp_rec, ",")
		);
		receivers = tmp_rec.str();
        receivers.erase(receivers.find_last_not_of(",") + 1);//delete last ','
	}
}

void Receivers::get_alarm_list_db(const vector<string> &handler_list, vector<string> &al_list)
{
	//DEBUG_STREAM << __func__ << ": entering";
	al_list.clear();
	for (auto ith : handler_list)
	{
		string dev_name = AlarmStringUtils::get_only_attr_name(ith);
		string tango_host = AlarmStringUtils::get_only_tango_host(ith);
		string host;
		int port;
		AlarmStringUtils::parse_tango_host(tango_host, host, port);
		//DEBUG_STREAM << __func__ << ": reading configuration from properties of " << ith << " -> " << dev_name << " on DB " << host << " : " << port;
		vector<string> att_list;

		//Tango::DbDevice *db_dev = get_db_device();
		Tango::Database *db = new Tango::Database(host, port);
		//db_dev->get_dbase()->get_device_attribute_list(dev_name,att_list);
		Tango::DbData db_data;
		db->get_device_attribute_list(dev_name,att_list);
		vector<string>::iterator it;
		for(it = att_list.begin(); it!=att_list.end(); it++)
		{
			//DEBUG_STREAM << __func__ << ": in " << dev_name << " found attribute " << *it;
			db_data.push_back(Tango::DbDatum(*it));
		}

		try
		{
			db->get_device_attribute_property(dev_name,db_data);
		}
		catch(Tango::DevFailed &e)
		{
			INFO_STREAM << __func__ << ": Exception reading configuration = " << e.errors[0].desc<<endl;
		}
		//DEBUG_STREAM << __func__ << ": in " << dev_name << " read  " << db_data.size() << " attributes with properties";
		for (size_t i=0;i < db_data.size();/*i++*/)
		{
			Tango::DevLong64 nb_prop;
			string &att_name = db_data[i].name;
			db_data[i] >> nb_prop;
			i++;
			string alm_name;
			string alm_receivers;
			string alm_groups;
			//DEBUG_STREAM << __func__ << ": in " << dev_name << " in attribute  " << att_name << " found " << nb_prop << " attribute properties db_data.size="<<db_data.size();
			for (long k=0;k < nb_prop;k++)
			{
				string &prop_name = db_data[i].name;

				if (prop_name == NAME_KEY)
					db_data[i] >> alm_name;
				else if (prop_name == RECEIVERS_KEY)
					db_data[i] >> alm_receivers;
				else if (prop_name == GROUPS_KEY)
					db_data[i] >> alm_groups;
				else
				{
					//DEBUG_STREAM << "att_name="<<att_name<<" UNUSED prop_name="<<prop_name<<endl;
					i++;
					continue;
				}
				i++;
			}
			if(alm_name.empty())
				continue;
			if(!alm_receivers.empty())
				DEBUG_STREAM << __func__ << ": in " << dev_name << " in attribute  " << att_name << " found receivers: '" << alm_receivers << "'";
			state_addr_map_t tmp_rec;
			AlarmStringUtils::parse_receivers(alm_receivers,tmp_rec);
			string alm_name_fqdn=ith+"/"+alm_name;
			al_list.push_back(alm_name_fqdn);
			//DEBUG_STREAM << __func__ << ": inserting map receivers alm="<<alm_name_fqdn << " -> receivers="<<alm_receivers<<endl;
			//receivers.insert(make_pair(alm_name_fqdn,tmp_rec));
			receiverlock.writerIn();
			receiverMap[alm_name_fqdn] = tmp_rec;
			groupMap[alm_name_fqdn] = alm_groups;
			receiverlock.writerOut();
		}
	}
	//DEBUG_STREAM << __func__ << ": exiting";
}

receivers_t Receivers::get_receivers(const string &name, const string &state)
{
	ReaderLock lock(receiverlock);
#if 0
	for(auto it : receiverMap)
	{
		cout << __func__ << ": name=" << it.first << endl;
		for(auto it2 : it.second)
		{
			cout << __func__ << ": .... state=" << it2.first << endl;
			for(auto it3 : it2.second.emails)
			{
				cout << __func__ << ": ........ dest=" << it3 << endl;
			}
		}
	}
#endif
	return receiverMap.at(name).at(state);
}

state_addr_map_t Receivers::get_state_addr(const string &name)
{
	ReaderLock lock(receiverlock);
	return receiverMap.at(name);
}

string Receivers::get_groups(const string &name)
{
	ReaderLock lock(receiverlock);
	return groupMap.at(name);
}

void Receivers::print_receivers(const string &name, string &out)
{
	ReaderLock lock(receiverlock);
	AlarmStringUtils::print_receivers(out, receiverMap.at(name));
}


void AlarmNotify::_remove_attribute(size_t idx, bool running, bool stopped)
{
	{
		std::lock_guard<std::mutex> lk(attributes_mutex);
		// Update the counters
		if(running)
			AttributeStartedNumber--;
		if(stopped)
			AttributeStoppedNumber--;

		AttributeNumber--;
	}
}

/*----- PROTECTED REGION END -----*/	//	AlarmNotify::namespace_ending
} //	namespace
